/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifndef HC_TIMER_H
#define HC_TIMER_H

void   hc_timer_set (hc_timer_t *a);
double hc_timer_get (hc_timer_t a);

#endif // HC_TIMER_H
