/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifdef KERNEL_STATIC
#include M2S(INCLUDE_PATH/inc_vendor.h)
#include M2S(INCLUDE_PATH/inc_types.h)
#include M2S(INCLUDE_PATH/inc_platform.cl)
#include M2S(INCLUDE_PATH/inc_common.cl)
#endif

KERNEL_FQ KERNEL_FA void m02000_m04 (KERN_ATTR_RULES ())
{
}

KERNEL_FQ KERNEL_FA void m02000_m08 (KERN_ATTR_RULES ())
{
}

KERNEL_FQ KERNEL_FA void m02000_m16 (KERN_ATTR_RULES ())
{
}

KERNEL_FQ KERNEL_FA void m02000_s04 (KERN_ATTR_RULES ())
{
}

KERNEL_FQ KERNEL_FA void m02000_s08 (KERN_ATTR_RULES ())
{
}

KERNEL_FQ KERNEL_FA void m02000_s16 (KERN_ATTR_RULES ())
{
}
