/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

//incompatible
//#define NEW_SIMD_CODE

#ifdef KERNEL_STATIC
#include M2S(INCLUDE_PATH/inc_vendor.h)
#include M2S(INCLUDE_PATH/inc_types.h)
#include M2S(INCLUDE_PATH/inc_platform.cl)
#include M2S(INCLUDE_PATH/inc_common.cl)
#include M2S(INCLUDE_PATH/inc_simd.cl)
#endif

KERNEL_FQ KERNEL_FA void m00200_m04 (KERN_ATTR_BASIC ())
{
  /**
   * modifier
   */

  const u64 lid = get_local_id (0);

  /**
   * base
   */

  const u64 gid = get_global_id (0);

  if (gid >= GID_CNT) return;

  u32 pw_buf0[4];
  u32 pw_buf1[4];

  pw_buf0[0] = pws[gid].i[0];
  pw_buf0[1] = pws[gid].i[1];
  pw_buf0[2] = pws[gid].i[2];
  pw_buf0[3] = pws[gid].i[3];
  pw_buf1[0] = pws[gid].i[4];
  pw_buf1[1] = pws[gid].i[5];
  pw_buf1[2] = pws[gid].i[6];
  pw_buf1[3] = pws[gid].i[7];

  const u32 pw_l_len = pws[gid].pw_len & 63;

  /**
   * loop
   */

  for (u32 il_pos = 0; il_pos < IL_CNT; il_pos += VECT_SIZE)
  {
    const u32x pw_r_len = pwlenx_create_combt (combs_buf, il_pos) & 63;

    const u32x pw_len = (pw_l_len + pw_r_len) & 63;

    /**
     * concat password candidate
     */

    u32x wordl0[4] = { 0 };
    u32x wordl1[4] = { 0 };
    u32x wordl2[4] = { 0 };
    u32x wordl3[4] = { 0 };

    wordl0[0] = pw_buf0[0];
    wordl0[1] = pw_buf0[1];
    wordl0[2] = pw_buf0[2];
    wordl0[3] = pw_buf0[3];
    wordl1[0] = pw_buf1[0];
    wordl1[1] = pw_buf1[1];
    wordl1[2] = pw_buf1[2];
    wordl1[3] = pw_buf1[3];

    u32x wordr0[4] = { 0 };
    u32x wordr1[4] = { 0 };
    u32x wordr2[4] = { 0 };
    u32x wordr3[4] = { 0 };

    wordr0[0] = ix_create_combt (combs_buf, il_pos, 0);
    wordr0[1] = ix_create_combt (combs_buf, il_pos, 1);
    wordr0[2] = ix_create_combt (combs_buf, il_pos, 2);
    wordr0[3] = ix_create_combt (combs_buf, il_pos, 3);
    wordr1[0] = ix_create_combt (combs_buf, il_pos, 4);
    wordr1[1] = ix_create_combt (combs_buf, il_pos, 5);
    wordr1[2] = ix_create_combt (combs_buf, il_pos, 6);
    wordr1[3] = ix_create_combt (combs_buf, il_pos, 7);

    if (COMBS_MODE == COMBINATOR_MODE_BASE_LEFT)
    {
      switch_buffer_by_offset_le_VV (wordr0, wordr1, wordr2, wordr3, pw_l_len);
    }
    else
    {
      switch_buffer_by_offset_le_VV (wordl0, wordl1, wordl2, wordl3, pw_r_len);
    }

    u32x w0[4];
    u32x w1[4];
    u32x w2[4];
    u32x w3[4];

    w0[0] = wordl0[0] | wordr0[0];
    w0[1] = wordl0[1] | wordr0[1];
    w0[2] = wordl0[2] | wordr0[2];
    w0[3] = wordl0[3] | wordr0[3];
    w1[0] = wordl1[0] | wordr1[0];
    w1[1] = wordl1[1] | wordr1[1];
    w1[2] = wordl1[2] | wordr1[2];
    w1[3] = wordl1[3] | wordr1[3];
    w2[0] = wordl2[0] | wordr2[0];
    w2[1] = wordl2[1] | wordr2[1];
    w2[2] = wordl2[2] | wordr2[2];
    w2[3] = wordl2[3] | wordr2[3];
    w3[0] = wordl3[0] | wordr3[0];
    w3[1] = wordl3[1] | wordr3[1];
    w3[2] = pw_len * 8;
    w3[3] = 0;

    /**
     * mysql
     */

    u32x w_t[16];

    w_t[ 0] = w0[0];
    w_t[ 1] = w0[1];
    w_t[ 2] = w0[2];
    w_t[ 3] = w0[3];
    w_t[ 4] = w1[0];
    w_t[ 5] = w1[1];
    w_t[ 6] = w1[2];
    w_t[ 7] = w1[3];
    w_t[ 8] = 0;
    w_t[ 9] = 0;
    w_t[10] = 0;
    w_t[11] = 0;
    w_t[12] = 0;
    w_t[13] = 0;
    w_t[14] = 0;
    w_t[15] = 0;

    u32x a = MYSQL323_A;
    u32x b = MYSQL323_B;
    u32x c = 0;
    u32x d = 0;

    u32x add = 7;

    #define ROUND(v)                              \
    {                                             \
      a ^= (((a & 0x3f) + add) * (v)) + (a << 8); \
      b += (b << 8) ^ a;                          \
      add += v;                                   \
    }

    int i;
    int j;

    for (i = 0, j = 0; i <= (int) pw_len - 4; i += 4, j += 1)
    {
      const u32x wj = w_t[j];

      ROUND (unpack_v8a_from_v32 (wj));
      ROUND (unpack_v8b_from_v32 (wj));
      ROUND (unpack_v8c_from_v32 (wj));
      ROUND (unpack_v8d_from_v32 (wj));
    }

    const u32x wj = w_t[j];

    const u32 left = pw_len - i;

    if (left == 3)
    {
      ROUND (unpack_v8a_from_v32 (wj));
      ROUND (unpack_v8b_from_v32 (wj));
      ROUND (unpack_v8c_from_v32 (wj));
    }
    else if (left == 2)
    {
      ROUND (unpack_v8a_from_v32 (wj));
      ROUND (unpack_v8b_from_v32 (wj));
    }
    else if (left == 1)
    {
      ROUND (unpack_v8a_from_v32 (wj));
    }

    a &= 0x7fffffff;
    b &= 0x7fffffff;

    u32x z = 0;

    COMPARE_M_SIMD (a, b, z, z);
  }
}

KERNEL_FQ KERNEL_FA void m00200_m08 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ KERNEL_FA void m00200_m16 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ KERNEL_FA void m00200_s04 (KERN_ATTR_BASIC ())
{
  /**
   * modifier
   */

  const u64 lid = get_local_id (0);

  /**
   * base
   */

  const u64 gid = get_global_id (0);

  if (gid >= GID_CNT) return;

  u32 pw_buf0[4];
  u32 pw_buf1[4];

  pw_buf0[0] = pws[gid].i[0];
  pw_buf0[1] = pws[gid].i[1];
  pw_buf0[2] = pws[gid].i[2];
  pw_buf0[3] = pws[gid].i[3];
  pw_buf1[0] = pws[gid].i[4];
  pw_buf1[1] = pws[gid].i[5];
  pw_buf1[2] = pws[gid].i[6];
  pw_buf1[3] = pws[gid].i[7];

  const u32 pw_l_len = pws[gid].pw_len & 63;

  /**
   * digest
   */

  const u32 search[4] =
  {
    digests_buf[DIGESTS_OFFSET_HOST].digest_buf[DGST_R0],
    digests_buf[DIGESTS_OFFSET_HOST].digest_buf[DGST_R1],
    0,
    0
  };

  /**
   * loop
   */

  for (u32 il_pos = 0; il_pos < IL_CNT; il_pos += VECT_SIZE)
  {
    const u32x pw_r_len = pwlenx_create_combt (combs_buf, il_pos) & 63;

    const u32x pw_len = (pw_l_len + pw_r_len) & 63;

    /**
     * concat password candidate
     */

    u32x wordl0[4] = { 0 };
    u32x wordl1[4] = { 0 };
    u32x wordl2[4] = { 0 };
    u32x wordl3[4] = { 0 };

    wordl0[0] = pw_buf0[0];
    wordl0[1] = pw_buf0[1];
    wordl0[2] = pw_buf0[2];
    wordl0[3] = pw_buf0[3];
    wordl1[0] = pw_buf1[0];
    wordl1[1] = pw_buf1[1];
    wordl1[2] = pw_buf1[2];
    wordl1[3] = pw_buf1[3];

    u32x wordr0[4] = { 0 };
    u32x wordr1[4] = { 0 };
    u32x wordr2[4] = { 0 };
    u32x wordr3[4] = { 0 };

    wordr0[0] = ix_create_combt (combs_buf, il_pos, 0);
    wordr0[1] = ix_create_combt (combs_buf, il_pos, 1);
    wordr0[2] = ix_create_combt (combs_buf, il_pos, 2);
    wordr0[3] = ix_create_combt (combs_buf, il_pos, 3);
    wordr1[0] = ix_create_combt (combs_buf, il_pos, 4);
    wordr1[1] = ix_create_combt (combs_buf, il_pos, 5);
    wordr1[2] = ix_create_combt (combs_buf, il_pos, 6);
    wordr1[3] = ix_create_combt (combs_buf, il_pos, 7);

    if (COMBS_MODE == COMBINATOR_MODE_BASE_LEFT)
    {
      switch_buffer_by_offset_le_VV (wordr0, wordr1, wordr2, wordr3, pw_l_len);
    }
    else
    {
      switch_buffer_by_offset_le_VV (wordl0, wordl1, wordl2, wordl3, pw_r_len);
    }

    u32x w0[4];
    u32x w1[4];
    u32x w2[4];
    u32x w3[4];

    w0[0] = wordl0[0] | wordr0[0];
    w0[1] = wordl0[1] | wordr0[1];
    w0[2] = wordl0[2] | wordr0[2];
    w0[3] = wordl0[3] | wordr0[3];
    w1[0] = wordl1[0] | wordr1[0];
    w1[1] = wordl1[1] | wordr1[1];
    w1[2] = wordl1[2] | wordr1[2];
    w1[3] = wordl1[3] | wordr1[3];
    w2[0] = wordl2[0] | wordr2[0];
    w2[1] = wordl2[1] | wordr2[1];
    w2[2] = wordl2[2] | wordr2[2];
    w2[3] = wordl2[3] | wordr2[3];
    w3[0] = wordl3[0] | wordr3[0];
    w3[1] = wordl3[1] | wordr3[1];
    w3[2] = pw_len * 8;
    w3[3] = 0;

    /**
     * mysql
     */

    u32x w_t[16];

    w_t[ 0] = w0[0];
    w_t[ 1] = w0[1];
    w_t[ 2] = w0[2];
    w_t[ 3] = w0[3];
    w_t[ 4] = w1[0];
    w_t[ 5] = w1[1];
    w_t[ 6] = w1[2];
    w_t[ 7] = w1[3];
    w_t[ 8] = 0;
    w_t[ 9] = 0;
    w_t[10] = 0;
    w_t[11] = 0;
    w_t[12] = 0;
    w_t[13] = 0;
    w_t[14] = 0;
    w_t[15] = 0;

    u32x a = MYSQL323_A;
    u32x b = MYSQL323_B;
    u32x c = 0;
    u32x d = 0;

    u32x add = 7;

    #define ROUND(v)                              \
    {                                             \
      a ^= (((a & 0x3f) + add) * (v)) + (a << 8); \
      b += (b << 8) ^ a;                          \
      add += v;                                   \
    }

    int i;
    int j;

    for (i = 0, j = 0; i <= (int) pw_len - 4; i += 4, j += 1)
    {
      const u32x wj = w_t[j];

      ROUND (unpack_v8a_from_v32 (wj));
      ROUND (unpack_v8b_from_v32 (wj));
      ROUND (unpack_v8c_from_v32 (wj));
      ROUND (unpack_v8d_from_v32 (wj));
    }

    const u32x wj = w_t[j];

    const u32 left = pw_len - i;

    if (left == 3)
    {
      ROUND (unpack_v8a_from_v32 (wj));
      ROUND (unpack_v8b_from_v32 (wj));
      ROUND (unpack_v8c_from_v32 (wj));
    }
    else if (left == 2)
    {
      ROUND (unpack_v8a_from_v32 (wj));
      ROUND (unpack_v8b_from_v32 (wj));
    }
    else if (left == 1)
    {
      ROUND (unpack_v8a_from_v32 (wj));
    }

    a &= 0x7fffffff;
    b &= 0x7fffffff;

    u32x z = 0;

    COMPARE_S_SIMD (a, b, z, z);
  }
}

KERNEL_FQ KERNEL_FA void m00200_s08 (KERN_ATTR_BASIC ())
{
}

KERNEL_FQ KERNEL_FA void m00200_s16 (KERN_ATTR_BASIC ())
{
}
