/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifndef HC_EMU_INC_CIPHER_DES_H
#define HC_EMU_INC_CIPHER_DES_H

#include "emu_general.h"

#include "inc_vendor.h"
#include "inc_cipher_des.h"

extern u32a c_SPtrans[8][64];
extern u32a c_skb[8][64];

#endif // HC_EMU_INC_CIPHER_DES_H
