/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

//#define NEW_SIMD_CODE

#ifdef KERNEL_STATIC
#include M2S(INCLUDE_PATH/inc_vendor.h)
#include M2S(INCLUDE_PATH/inc_types.h)
#include M2S(INCLUDE_PATH/inc_platform.cl)
#include M2S(INCLUDE_PATH/inc_common.cl)
#include M2S(INCLUDE_PATH/inc_simd.cl)
#include M2S(INCLUDE_PATH/inc_hash_md5.cl)
#endif

typedef struct tacacs_plus
{
  u32 session_buf[16];

  u32 ct_data_buf[64];
  u32 ct_data_len;

  u32 sequence_buf[16];

} tacacs_plus_t;

DECLSPEC void m16100m (PRIVATE_AS u32 *w0, PRIVATE_AS u32 *w1, PRIVATE_AS u32 *w2, PRIVATE_AS u32 *w3, const u32 pw_len, KERN_ATTR_FUNC_ESALT (tacacs_plus_t))
{
  /**
   * modifiers are taken from args
   */

  /**
   * digest
   */

  const u32 ct_len = esalt_bufs[DIGESTS_OFFSET_HOST].ct_data_len;

  u32 ct_buf[2];

  ct_buf[0] = esalt_bufs[DIGESTS_OFFSET_HOST].ct_data_buf[0];
  ct_buf[1] = esalt_bufs[DIGESTS_OFFSET_HOST].ct_data_buf[1];

  /**
   * salt
   */

  const u32 session_buf  = esalt_bufs[DIGESTS_OFFSET_HOST].session_buf[0];
  const u32 sequence_buf = esalt_bufs[DIGESTS_OFFSET_HOST].sequence_buf[0];

  const u32 pw_salt_len = 4 + pw_len + 2;

  u32 s0[4];
  u32 s1[4];
  u32 s2[4];
  u32 s3[4];

  s0[0] = sequence_buf | 0x00800000;
  s0[1] = 0;
  s0[2] = 0;
  s0[3] = 0;
  s1[0] = 0;
  s1[1] = 0;
  s1[2] = 0;
  s1[3] = 0;
  s2[0] = 0;
  s2[1] = 0;
  s2[2] = 0;
  s2[3] = 0;
  s3[0] = 0;
  s3[1] = 0;
  s3[2] = 0;
  s3[3] = 0;

  switch_buffer_by_offset_le (s0, s1, s2, s3, 4 + pw_len);

  /**
   * loop
   */

  const u32 w0l = w0[0];

  for (u32 il_pos = 0; il_pos < IL_CNT; il_pos += VECT_SIZE)
  {
    const u32x w0r = ix_create_bft (bfs_buf, il_pos);

    const u32x w0lr = w0l | w0r;

    u32x t0[4];
    u32x t1[4];
    u32x t2[4];
    u32x t3[4];

    t0[0] = session_buf;
    t0[1] = w0lr;
    t0[2] = w0[1];
    t0[3] = w0[2];
    t1[0] = w0[3];
    t1[1] = w1[0];
    t1[2] = w1[1];
    t1[3] = w1[2];
    t2[0] = w1[3];
    t2[1] = w2[0];
    t2[2] = w2[1];
    t2[3] = w2[2];
    t3[0] = w2[3];
    t3[1] = w3[0];
    t3[2] = pw_salt_len * 8;
    t3[3] = 0;

    t0[1] |= s0[1];
    t0[2] |= s0[2];
    t0[3] |= s0[3];
    t1[0] |= s1[0];
    t1[1] |= s1[1];
    t1[2] |= s1[2];
    t1[3] |= s1[3];
    t2[0] |= s2[0];
    t2[1] |= s2[1];
    t2[2] |= s2[2];
    t2[3] |= s2[3];
    t3[0] |= s3[0];
    t3[1] |= s3[1];

    /**
     * md5
     */

    u32x a = MD5M_A;
    u32x b = MD5M_B;
    u32x c = MD5M_C;
    u32x d = MD5M_D;

    MD5_STEP (MD5_Fo, a, b, c, d, t0[0], MD5C00, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t0[1], MD5C01, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t0[2], MD5C02, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t0[3], MD5C03, MD5S03);
    MD5_STEP (MD5_Fo, a, b, c, d, t1[0], MD5C04, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t1[1], MD5C05, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t1[2], MD5C06, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t1[3], MD5C07, MD5S03);
    MD5_STEP (MD5_Fo, a, b, c, d, t2[0], MD5C08, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t2[1], MD5C09, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t2[2], MD5C0a, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t2[3], MD5C0b, MD5S03);
    MD5_STEP (MD5_Fo, a, b, c, d, t3[0], MD5C0c, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t3[1], MD5C0d, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t3[2], MD5C0e, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t3[3], MD5C0f, MD5S03);

    MD5_STEP (MD5_Go, a, b, c, d, t0[1], MD5C10, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t1[2], MD5C11, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t2[3], MD5C12, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t0[0], MD5C13, MD5S13);
    MD5_STEP (MD5_Go, a, b, c, d, t1[1], MD5C14, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t2[2], MD5C15, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t3[3], MD5C16, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t1[0], MD5C17, MD5S13);
    MD5_STEP (MD5_Go, a, b, c, d, t2[1], MD5C18, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t3[2], MD5C19, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t0[3], MD5C1a, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t2[0], MD5C1b, MD5S13);
    MD5_STEP (MD5_Go, a, b, c, d, t3[1], MD5C1c, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t0[2], MD5C1d, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t1[3], MD5C1e, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t3[0], MD5C1f, MD5S13);

    u32x t;

    MD5_STEP (MD5_H1, a, b, c, d, t1[1], MD5C20, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t2[0], MD5C21, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t2[3], MD5C22, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t3[2], MD5C23, MD5S23);
    MD5_STEP (MD5_H1, a, b, c, d, t0[1], MD5C24, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t1[0], MD5C25, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t1[3], MD5C26, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t2[2], MD5C27, MD5S23);
    MD5_STEP (MD5_H1, a, b, c, d, t3[1], MD5C28, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t0[0], MD5C29, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t0[3], MD5C2a, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t1[2], MD5C2b, MD5S23);
    MD5_STEP (MD5_H1, a, b, c, d, t2[1], MD5C2c, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t3[0], MD5C2d, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t3[3], MD5C2e, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t0[2], MD5C2f, MD5S23);

    MD5_STEP (MD5_I , a, b, c, d, t0[0], MD5C30, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t1[3], MD5C31, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t3[2], MD5C32, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t1[1], MD5C33, MD5S33);
    MD5_STEP (MD5_I , a, b, c, d, t3[0], MD5C34, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t0[3], MD5C35, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t2[2], MD5C36, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t0[1], MD5C37, MD5S33);
    MD5_STEP (MD5_I , a, b, c, d, t2[0], MD5C38, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t3[3], MD5C39, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t1[2], MD5C3a, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t3[1], MD5C3b, MD5S33);
    MD5_STEP (MD5_I , a, b, c, d, t1[0], MD5C3c, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t2[3], MD5C3d, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t0[2], MD5C3e, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t2[1], MD5C3f, MD5S33);

    a += MD5M_A;
    b += MD5M_B;
    c += MD5M_C;
    d += MD5M_D;

    u32 test[2];

    test[0] = a ^ ct_buf[0];
    test[1] = b ^ ct_buf[1];

    if (sequence_buf == 0x01c0)
    {
      const u32 action          = ((test[0] >>  0) & 0xff);
      // can have more than predefined ones
      // const u32 priv_lvl        = ((test[0] >>  8) & 0xff);
      const u32 authen_type     = ((test[0] >> 16) & 0xff);
      const u32 authen_service  = ((test[0] >> 24) & 0xff);
      const u32 user_len        = ((test[1] >>  0) & 0xff);
      const u32 port_len        = ((test[1] >>  8) & 0xff);
      const u32 rem_addr_len    = ((test[1] >> 16) & 0xff);
      const u32 data_len        = ((test[1] >> 24) & 0xff);

      if (((action == 0x01) || (action == 0x02) || (action == 0x04))
      &&  ((authen_type    >= 0x01) && (authen_type    <= 0x06))
      &&  ((authen_service >= 0x00) && (authen_service <= 0x09))
      &&  ((8 + user_len + port_len + rem_addr_len + data_len) == ct_len))
      {
        if (hc_atomic_inc (&hashes_shown[DIGESTS_OFFSET_HOST]) == 0)
        {
          mark_hash (plains_buf, d_return_buf, SALT_POS_HOST, DIGESTS_CNT, 0, DIGESTS_OFFSET_HOST + 0, gid, il_pos, 0, 0);
        }
      }
    }
    else if ((sequence_buf == 0x03c0) || (sequence_buf == 0x05c0))
    {
      const u32 msg_len   = ((test[0] >>  0) & 0xff) << 8
                          | ((test[0] >>  8) & 0xff) << 0;
      const u32 data_len  = ((test[0] >> 16) & 0xff) << 8
                          | ((test[0] >> 24) & 0xff) << 0;
      const u32 flags     = ((test[1] >>  0) & 0xff);

      if (((5 + msg_len) == ct_len)
       &&  (data_len == 0)
       &&  (flags == 0))
      {
        if (hc_atomic_inc (&hashes_shown[DIGESTS_OFFSET_HOST]) == 0)
        {
          mark_hash (plains_buf, d_return_buf, SALT_POS_HOST, DIGESTS_CNT, 0, DIGESTS_OFFSET_HOST + 0, gid, il_pos, 0, 0);
        }
      }
    }
    else
    {
      const u32 status    = ((test[0] >>  0) & 0xff);
      const u32 flags     = ((test[0] >>  8) & 0xff);
      const u32 msg_len   = ((test[0] >> 16) & 0xff) << 8
                          | ((test[0] >> 24) & 0xff) << 0;
      const u32 data_len  = ((test[1] >>  0) & 0xff) << 8
                          | ((test[1] >>  8) & 0xff) << 0;

      if (((status >= 0x01 && status <= 0x07) || status == 0x21)
       &&  (flags == 0x01 || flags == 0x00)
       &&  (6 + msg_len + data_len == ct_len))
      {
        if (hc_atomic_inc (&hashes_shown[DIGESTS_OFFSET_HOST]) == 0)
        {
          mark_hash (plains_buf, d_return_buf, SALT_POS_HOST, DIGESTS_CNT, 0, DIGESTS_OFFSET_HOST + 0, gid, il_pos, 0, 0);
        }
      }
    }
  }
}

DECLSPEC void m16100s (PRIVATE_AS u32 *w0, PRIVATE_AS u32 *w1, PRIVATE_AS u32 *w2, PRIVATE_AS u32 *w3, const u32 pw_len, KERN_ATTR_FUNC_ESALT (tacacs_plus_t))
{
  /**
   * modifiers are taken from args
   */

  /**
   * digest
   */

  const u32 ct_len = esalt_bufs[DIGESTS_OFFSET_HOST].ct_data_len;

  u32 ct_buf[2];

  ct_buf[0] = esalt_bufs[DIGESTS_OFFSET_HOST].ct_data_buf[0];
  ct_buf[1] = esalt_bufs[DIGESTS_OFFSET_HOST].ct_data_buf[1];

  /**
   * salt
   */

  const u32 session_buf  = esalt_bufs[DIGESTS_OFFSET_HOST].session_buf[0];
  const u32 sequence_buf = esalt_bufs[DIGESTS_OFFSET_HOST].sequence_buf[0];

  const u32 pw_salt_len = 4 + pw_len + 2;

  u32 s0[4];
  u32 s1[4];
  u32 s2[4];
  u32 s3[4];

  s0[0] = sequence_buf | 0x00800000;
  s0[1] = 0;
  s0[2] = 0;
  s0[3] = 0;
  s1[0] = 0;
  s1[1] = 0;
  s1[2] = 0;
  s1[3] = 0;
  s2[0] = 0;
  s2[1] = 0;
  s2[2] = 0;
  s2[3] = 0;
  s3[0] = 0;
  s3[1] = 0;
  s3[2] = 0;
  s3[3] = 0;

  switch_buffer_by_offset_le (s0, s1, s2, s3, 4 + pw_len);

  /**
   * loop
   */

  const u32 w0l = w0[0];

  for (u32 il_pos = 0; il_pos < IL_CNT; il_pos += VECT_SIZE)
  {
    const u32x w0r = ix_create_bft (bfs_buf, il_pos);

    const u32x w0lr = w0l | w0r;

    u32x t0[4];
    u32x t1[4];
    u32x t2[4];
    u32x t3[4];

    t0[0] = session_buf;
    t0[1] = w0lr;
    t0[2] = w0[1];
    t0[3] = w0[2];
    t1[0] = w0[3];
    t1[1] = w1[0];
    t1[2] = w1[1];
    t1[3] = w1[2];
    t2[0] = w1[3];
    t2[1] = w2[0];
    t2[2] = w2[1];
    t2[3] = w2[2];
    t3[0] = w2[3];
    t3[1] = w3[0];
    t3[2] = pw_salt_len * 8;
    t3[3] = 0;

    t0[1] |= s0[1];
    t0[2] |= s0[2];
    t0[3] |= s0[3];
    t1[0] |= s1[0];
    t1[1] |= s1[1];
    t1[2] |= s1[2];
    t1[3] |= s1[3];
    t2[0] |= s2[0];
    t2[1] |= s2[1];
    t2[2] |= s2[2];
    t2[3] |= s2[3];
    t3[0] |= s3[0];
    t3[1] |= s3[1];

    /**
     * md5
     */

    u32x a = MD5M_A;
    u32x b = MD5M_B;
    u32x c = MD5M_C;
    u32x d = MD5M_D;

    MD5_STEP (MD5_Fo, a, b, c, d, t0[0], MD5C00, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t0[1], MD5C01, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t0[2], MD5C02, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t0[3], MD5C03, MD5S03);
    MD5_STEP (MD5_Fo, a, b, c, d, t1[0], MD5C04, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t1[1], MD5C05, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t1[2], MD5C06, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t1[3], MD5C07, MD5S03);
    MD5_STEP (MD5_Fo, a, b, c, d, t2[0], MD5C08, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t2[1], MD5C09, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t2[2], MD5C0a, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t2[3], MD5C0b, MD5S03);
    MD5_STEP (MD5_Fo, a, b, c, d, t3[0], MD5C0c, MD5S00);
    MD5_STEP (MD5_Fo, d, a, b, c, t3[1], MD5C0d, MD5S01);
    MD5_STEP (MD5_Fo, c, d, a, b, t3[2], MD5C0e, MD5S02);
    MD5_STEP (MD5_Fo, b, c, d, a, t3[3], MD5C0f, MD5S03);

    MD5_STEP (MD5_Go, a, b, c, d, t0[1], MD5C10, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t1[2], MD5C11, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t2[3], MD5C12, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t0[0], MD5C13, MD5S13);
    MD5_STEP (MD5_Go, a, b, c, d, t1[1], MD5C14, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t2[2], MD5C15, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t3[3], MD5C16, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t1[0], MD5C17, MD5S13);
    MD5_STEP (MD5_Go, a, b, c, d, t2[1], MD5C18, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t3[2], MD5C19, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t0[3], MD5C1a, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t2[0], MD5C1b, MD5S13);
    MD5_STEP (MD5_Go, a, b, c, d, t3[1], MD5C1c, MD5S10);
    MD5_STEP (MD5_Go, d, a, b, c, t0[2], MD5C1d, MD5S11);
    MD5_STEP (MD5_Go, c, d, a, b, t1[3], MD5C1e, MD5S12);
    MD5_STEP (MD5_Go, b, c, d, a, t3[0], MD5C1f, MD5S13);

    u32x t;

    MD5_STEP (MD5_H1, a, b, c, d, t1[1], MD5C20, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t2[0], MD5C21, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t2[3], MD5C22, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t3[2], MD5C23, MD5S23);
    MD5_STEP (MD5_H1, a, b, c, d, t0[1], MD5C24, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t1[0], MD5C25, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t1[3], MD5C26, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t2[2], MD5C27, MD5S23);
    MD5_STEP (MD5_H1, a, b, c, d, t3[1], MD5C28, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t0[0], MD5C29, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t0[3], MD5C2a, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t1[2], MD5C2b, MD5S23);
    MD5_STEP (MD5_H1, a, b, c, d, t2[1], MD5C2c, MD5S20);
    MD5_STEP (MD5_H2, d, a, b, c, t3[0], MD5C2d, MD5S21);
    MD5_STEP (MD5_H1, c, d, a, b, t3[3], MD5C2e, MD5S22);
    MD5_STEP (MD5_H2, b, c, d, a, t0[2], MD5C2f, MD5S23);

    MD5_STEP (MD5_I , a, b, c, d, t0[0], MD5C30, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t1[3], MD5C31, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t3[2], MD5C32, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t1[1], MD5C33, MD5S33);
    MD5_STEP (MD5_I , a, b, c, d, t3[0], MD5C34, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t0[3], MD5C35, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t2[2], MD5C36, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t0[1], MD5C37, MD5S33);
    MD5_STEP (MD5_I , a, b, c, d, t2[0], MD5C38, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t3[3], MD5C39, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t1[2], MD5C3a, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t3[1], MD5C3b, MD5S33);
    MD5_STEP (MD5_I , a, b, c, d, t1[0], MD5C3c, MD5S30);
    MD5_STEP (MD5_I , d, a, b, c, t2[3], MD5C3d, MD5S31);
    MD5_STEP (MD5_I , c, d, a, b, t0[2], MD5C3e, MD5S32);
    MD5_STEP (MD5_I , b, c, d, a, t2[1], MD5C3f, MD5S33);

    a += MD5M_A;
    b += MD5M_B;
    c += MD5M_C;
    d += MD5M_D;

    u32 test[2];

    test[0] = a ^ ct_buf[0];
    test[1] = b ^ ct_buf[1];

    if (sequence_buf == 0x01c0)
    {
      const u32 action          = ((test[0] >>  0) & 0xff);
      // can have more than predefined ones
      // const u32 priv_lvl        = ((test[0] >>  8) & 0xff);
      const u32 authen_type     = ((test[0] >> 16) & 0xff);
      const u32 authen_service  = ((test[0] >> 24) & 0xff);
      const u32 user_len        = ((test[1] >>  0) & 0xff);
      const u32 port_len        = ((test[1] >>  8) & 0xff);
      const u32 rem_addr_len    = ((test[1] >> 16) & 0xff);
      const u32 data_len        = ((test[1] >> 24) & 0xff);

      if (((action == 0x01) || (action == 0x02) || (action == 0x04))
      &&  ((authen_type    >= 0x01) && (authen_type    <= 0x06))
      &&  ((authen_service >= 0x00) && (authen_service <= 0x09))
      &&  ((8 + user_len + port_len + rem_addr_len + data_len) == ct_len))
      {
        if (hc_atomic_inc (&hashes_shown[DIGESTS_OFFSET_HOST]) == 0)
        {
          mark_hash (plains_buf, d_return_buf, SALT_POS_HOST, DIGESTS_CNT, 0, DIGESTS_OFFSET_HOST + 0, gid, il_pos, 0, 0);
        }
      }
    }
    else if ((sequence_buf == 0x03c0) || (sequence_buf == 0x05c0))
    {
      const u32 msg_len   = ((test[0] >>  0) & 0xff) << 8
                          | ((test[0] >>  8) & 0xff) << 0;
      const u32 data_len  = ((test[0] >> 16) & 0xff) << 8
                          | ((test[0] >> 24) & 0xff) << 0;
      const u32 flags     = ((test[1] >>  0) & 0xff);

      if (((5 + msg_len) == ct_len)
       &&  (data_len == 0)
       &&  (flags == 0))
      {
        if (hc_atomic_inc (&hashes_shown[DIGESTS_OFFSET_HOST]) == 0)
        {
          mark_hash (plains_buf, d_return_buf, SALT_POS_HOST, DIGESTS_CNT, 0, DIGESTS_OFFSET_HOST + 0, gid, il_pos, 0, 0);
        }
      }
    }
    else
    {
      const u32 status    = ((test[0] >>  0) & 0xff);
      const u32 flags     = ((test[0] >>  8) & 0xff);
      const u32 msg_len   = ((test[0] >> 16) & 0xff) << 8
                          | ((test[0] >> 24) & 0xff) << 0;
      const u32 data_len  = ((test[1] >>  0) & 0xff) << 8
                          | ((test[1] >>  8) & 0xff) << 0;

      if (((status >= 0x01 && status <= 0x07) || status == 0x21)
       &&  (flags == 0x01 || flags == 0x00)
       &&  (6 + msg_len + data_len == ct_len))
      {
        if (hc_atomic_inc (&hashes_shown[DIGESTS_OFFSET_HOST]) == 0)
        {
          mark_hash (plains_buf, d_return_buf, SALT_POS_HOST, DIGESTS_CNT, 0, DIGESTS_OFFSET_HOST + 0, gid, il_pos, 0, 0);
        }
      }
    }
  }
}

KERNEL_FQ KERNEL_FA void m16100_m04 (KERN_ATTR_ESALT (tacacs_plus_t))
{
  /**
   * base
   */

  const u64 lid = get_local_id (0);
  const u64 gid = get_global_id (0);
  const u64 lsz = get_local_size (0);

  if (gid >= GID_CNT) return;

  u32 w0[4];

  w0[0] = pws[gid].i[ 0];
  w0[1] = pws[gid].i[ 1];
  w0[2] = pws[gid].i[ 2];
  w0[3] = pws[gid].i[ 3];

  u32 w1[4];

  w1[0] = 0;
  w1[1] = 0;
  w1[2] = 0;
  w1[3] = 0;

  u32 w2[4];

  w2[0] = 0;
  w2[1] = 0;
  w2[2] = 0;
  w2[3] = 0;

  u32 w3[4];

  w3[0] = 0;
  w3[1] = 0;
  w3[2] = 0;
  w3[3] = 0;

  const u32 pw_len = pws[gid].pw_len & 63;

  /**
   * main
   */

  m16100m (w0, w1, w2, w3, pw_len, pws, rules_buf, combs_buf, bfs_buf, tmps, hooks, bitmaps_buf_s1_a, bitmaps_buf_s1_b, bitmaps_buf_s1_c, bitmaps_buf_s1_d, bitmaps_buf_s2_a, bitmaps_buf_s2_b, bitmaps_buf_s2_c, bitmaps_buf_s2_d, plains_buf, digests_buf, hashes_shown, salt_bufs, esalt_bufs, d_return_buf, d_extra0_buf, d_extra1_buf, d_extra2_buf, d_extra3_buf, kernel_param, gid, lid, lsz);
}

KERNEL_FQ KERNEL_FA void m16100_m08 (KERN_ATTR_ESALT (tacacs_plus_t))
{
  /**
   * base
   */

  const u64 lid = get_local_id (0);
  const u64 gid = get_global_id (0);
  const u64 lsz = get_local_size (0);

  if (gid >= GID_CNT) return;

  u32 w0[4];

  w0[0] = pws[gid].i[ 0];
  w0[1] = pws[gid].i[ 1];
  w0[2] = pws[gid].i[ 2];
  w0[3] = pws[gid].i[ 3];

  u32 w1[4];

  w1[0] = pws[gid].i[ 4];
  w1[1] = pws[gid].i[ 5];
  w1[2] = pws[gid].i[ 6];
  w1[3] = pws[gid].i[ 7];

  u32 w2[4];

  w2[0] = 0;
  w2[1] = 0;
  w2[2] = 0;
  w2[3] = 0;

  u32 w3[4];

  w3[0] = 0;
  w3[1] = 0;
  w3[2] = 0;
  w3[3] = 0;

  const u32 pw_len = pws[gid].pw_len & 63;

  /**
   * main
   */

  m16100m (w0, w1, w2, w3, pw_len, pws, rules_buf, combs_buf, bfs_buf, tmps, hooks, bitmaps_buf_s1_a, bitmaps_buf_s1_b, bitmaps_buf_s1_c, bitmaps_buf_s1_d, bitmaps_buf_s2_a, bitmaps_buf_s2_b, bitmaps_buf_s2_c, bitmaps_buf_s2_d, plains_buf, digests_buf, hashes_shown, salt_bufs, esalt_bufs, d_return_buf, d_extra0_buf, d_extra1_buf, d_extra2_buf, d_extra3_buf, kernel_param, gid, lid, lsz);
}

KERNEL_FQ KERNEL_FA void m16100_m16 (KERN_ATTR_ESALT (tacacs_plus_t))
{
  /**
   * base
   */

  const u64 lid = get_local_id (0);
  const u64 gid = get_global_id (0);
  const u64 lsz = get_local_size (0);

  if (gid >= GID_CNT) return;

  u32 w0[4];

  w0[0] = pws[gid].i[ 0];
  w0[1] = pws[gid].i[ 1];
  w0[2] = pws[gid].i[ 2];
  w0[3] = pws[gid].i[ 3];

  u32 w1[4];

  w1[0] = pws[gid].i[ 4];
  w1[1] = pws[gid].i[ 5];
  w1[2] = pws[gid].i[ 6];
  w1[3] = pws[gid].i[ 7];

  u32 w2[4];

  w2[0] = pws[gid].i[ 8];
  w2[1] = pws[gid].i[ 9];
  w2[2] = pws[gid].i[10];
  w2[3] = pws[gid].i[11];

  u32 w3[4];

  w3[0] = pws[gid].i[12];
  w3[1] = pws[gid].i[13];
  w3[2] = 0;
  w3[3] = 0;

  const u32 pw_len = pws[gid].pw_len & 63;

  /**
   * main
   */

  m16100m (w0, w1, w2, w3, pw_len, pws, rules_buf, combs_buf, bfs_buf, tmps, hooks, bitmaps_buf_s1_a, bitmaps_buf_s1_b, bitmaps_buf_s1_c, bitmaps_buf_s1_d, bitmaps_buf_s2_a, bitmaps_buf_s2_b, bitmaps_buf_s2_c, bitmaps_buf_s2_d, plains_buf, digests_buf, hashes_shown, salt_bufs, esalt_bufs, d_return_buf, d_extra0_buf, d_extra1_buf, d_extra2_buf, d_extra3_buf, kernel_param, gid, lid, lsz);
}

KERNEL_FQ KERNEL_FA void m16100_s04 (KERN_ATTR_ESALT (tacacs_plus_t))
{
  /**
   * base
   */

  const u64 lid = get_local_id (0);
  const u64 gid = get_global_id (0);
  const u64 lsz = get_local_size (0);

  if (gid >= GID_CNT) return;

  u32 w0[4];

  w0[0] = pws[gid].i[ 0];
  w0[1] = pws[gid].i[ 1];
  w0[2] = pws[gid].i[ 2];
  w0[3] = pws[gid].i[ 3];

  u32 w1[4];

  w1[0] = 0;
  w1[1] = 0;
  w1[2] = 0;
  w1[3] = 0;

  u32 w2[4];

  w2[0] = 0;
  w2[1] = 0;
  w2[2] = 0;
  w2[3] = 0;

  u32 w3[4];

  w3[0] = 0;
  w3[1] = 0;
  w3[2] = 0;
  w3[3] = 0;

  const u32 pw_len = pws[gid].pw_len & 63;

  /**
   * main
   */

  m16100s (w0, w1, w2, w3, pw_len, pws, rules_buf, combs_buf, bfs_buf, tmps, hooks, bitmaps_buf_s1_a, bitmaps_buf_s1_b, bitmaps_buf_s1_c, bitmaps_buf_s1_d, bitmaps_buf_s2_a, bitmaps_buf_s2_b, bitmaps_buf_s2_c, bitmaps_buf_s2_d, plains_buf, digests_buf, hashes_shown, salt_bufs, esalt_bufs, d_return_buf, d_extra0_buf, d_extra1_buf, d_extra2_buf, d_extra3_buf, kernel_param, gid, lid, lsz);
}

KERNEL_FQ KERNEL_FA void m16100_s08 (KERN_ATTR_ESALT (tacacs_plus_t))
{
  /**
   * base
   */

  const u64 lid = get_local_id (0);
  const u64 gid = get_global_id (0);
  const u64 lsz = get_local_size (0);

  if (gid >= GID_CNT) return;

  u32 w0[4];

  w0[0] = pws[gid].i[ 0];
  w0[1] = pws[gid].i[ 1];
  w0[2] = pws[gid].i[ 2];
  w0[3] = pws[gid].i[ 3];

  u32 w1[4];

  w1[0] = pws[gid].i[ 4];
  w1[1] = pws[gid].i[ 5];
  w1[2] = pws[gid].i[ 6];
  w1[3] = pws[gid].i[ 7];

  u32 w2[4];

  w2[0] = 0;
  w2[1] = 0;
  w2[2] = 0;
  w2[3] = 0;

  u32 w3[4];

  w3[0] = 0;
  w3[1] = 0;
  w3[2] = 0;
  w3[3] = 0;

  const u32 pw_len = pws[gid].pw_len & 63;

  /**
   * main
   */

  m16100s (w0, w1, w2, w3, pw_len, pws, rules_buf, combs_buf, bfs_buf, tmps, hooks, bitmaps_buf_s1_a, bitmaps_buf_s1_b, bitmaps_buf_s1_c, bitmaps_buf_s1_d, bitmaps_buf_s2_a, bitmaps_buf_s2_b, bitmaps_buf_s2_c, bitmaps_buf_s2_d, plains_buf, digests_buf, hashes_shown, salt_bufs, esalt_bufs, d_return_buf, d_extra0_buf, d_extra1_buf, d_extra2_buf, d_extra3_buf, kernel_param, gid, lid, lsz);
}

KERNEL_FQ KERNEL_FA void m16100_s16 (KERN_ATTR_ESALT (tacacs_plus_t))
{
  /**
   * base
   */

  const u64 lid = get_local_id (0);
  const u64 gid = get_global_id (0);
  const u64 lsz = get_local_size (0);

  if (gid >= GID_CNT) return;

  u32 w0[4];

  w0[0] = pws[gid].i[ 0];
  w0[1] = pws[gid].i[ 1];
  w0[2] = pws[gid].i[ 2];
  w0[3] = pws[gid].i[ 3];

  u32 w1[4];

  w1[0] = pws[gid].i[ 4];
  w1[1] = pws[gid].i[ 5];
  w1[2] = pws[gid].i[ 6];
  w1[3] = pws[gid].i[ 7];

  u32 w2[4];

  w2[0] = pws[gid].i[ 8];
  w2[1] = pws[gid].i[ 9];
  w2[2] = pws[gid].i[10];
  w2[3] = pws[gid].i[11];

  u32 w3[4];

  w3[0] = pws[gid].i[12];
  w3[1] = pws[gid].i[13];
  w3[2] = 0;
  w3[3] = 0;

  const u32 pw_len = pws[gid].pw_len & 63;

  /**
   * main
   */

  m16100s (w0, w1, w2, w3, pw_len, pws, rules_buf, combs_buf, bfs_buf, tmps, hooks, bitmaps_buf_s1_a, bitmaps_buf_s1_b, bitmaps_buf_s1_c, bitmaps_buf_s1_d, bitmaps_buf_s2_a, bitmaps_buf_s2_b, bitmaps_buf_s2_c, bitmaps_buf_s2_d, plains_buf, digests_buf, hashes_shown, salt_bufs, esalt_bufs, d_return_buf, d_extra0_buf, d_extra1_buf, d_extra2_buf, d_extra3_buf, kernel_param, gid, lid, lsz);
}
